/*
 * Procesor de efecte pentru chitara
 * Constantin Mihalache <mihalache.c94@gmail.com>
 * Grupa 333CB
 */
#define SGN_MIN -512
#define SGN_MAX 512
#define IN_WAKEUP 10
#define TIMER_1HZ 15625
#define MAX_OUTPUT_HISTORY 800

#include "lcd.h"
#include "menu.h"
#include "twi.h"

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <util/atomic.h>
#include <math.h>

char buffer[17];
uint8_t ui_check = 0;
// if active_efects & _BV(x) != 0 => x effect is active
uint16_t active_effects;

float normalized_output;
int16_t output_dac;
int16_t read_adc;

int16_t output_history[MAX_OUTPUT_HISTORY];
int16_t output_history_last = 0;
int16_t output_history_first = 1;

#define OUTPUT_HISTORY_ADD(x) do { \
    output_history[output_history_last] = x; \
    output_history_last = (output_history_last + 1) % MAX_OUTPUT_HISTORY; \
    output_history_first = (output_history_first + 1) % MAX_OUTPUT_HISTORY; \
    } while(0);

uint16_t timer0_counter = 0;
uint16_t timer0_top = 0;

#define TCNT0E ((timer0_counter + TCNT0) % (timer0_top + 1))

ISR(TIMER0_COMPA_vect)
{
    timer0_counter = (timer0_counter + 0xFF) % (timer0_top + 1);
}

// Pin change interrupt handler
ISR(PCINT1_vect)
{
    ui_check = 1;
}

// keep signal in range [-SGN_MIN, SGN_MAX]
float adjust_amplitude(float x)
{
    if (x < -1.0f)
        return -1.0f;
    if (x > 1.0f)
        return 1.0f;
    return x;
}

float apply_distortion(float x)
{
    // TODO settings
    float drive = 0.8f;
    float k = (2 * drive) / (1 - drive);

    return ((1 + k) * x) / (1 + k * fabs(x));
}

float apply_tremolo(float x)
{
    // TODO settings
    float depth = 0.8f;
    float modulation = sin((2 * M_PI / OCR1A) * TCNT1);
    modulation = (1 - depth) + depth * modulation * modulation;

    return x * modulation;
}

float apply_flanger(float x)
{
    float delay = 0.01f;
    float amp = 0.5f;
    float max_sample_delay = delay * MAX_OUTPUT_HISTORY;
    float current_sin = fabs(sin((2 * M_PI / timer0_top) * TCNT0E));
    int16_t current_delay = (int16_t)ceil(current_sin * max_sample_delay);

    float out_delayed = (float)output_history[(output_history_last + (MAX_OUTPUT_HISTORY - 1) - current_delay) % MAX_OUTPUT_HISTORY];
    out_delayed /= SGN_MAX;

    return x * amp + 
        (1 - amp) * out_delayed;
}

float apply_echo(float x)
{
    // TODO linear interpolation between current sample and past sample
    return x;
}

float apply_compression(float x)
{
    // TODO
    return x;
}

float process_signal(float x)
{
    // x = apply_distortion(x);

    // x = apply_tremolo(x);

    // x = apply_flanger(x);

    x = apply_echo(x);

    // x = apply_compression(x);

    return adjust_amplitude(x);
}

// ADC conversion complete handler
ISR(ADC_vect)
{
    // gather from ADC
    read_adc = ADC;

    // Twos Complement converter
    if (read_adc & _BV(9)) {
        //negative number
        read_adc = read_adc | ~(_BV(10) - 1);
    }

    // process normalized signal
    output_dac = (int16_t)(process_signal(((float)read_adc) / SGN_MAX) * SGN_MAX);
    OUTPUT_HISTORY_ADD(output_dac);
    output_dac += SGN_MAX;

    DAC_output(output_dac);

    if (ui_check) {
        // check for menu key inputs
        if ((PINB & _BV(PB0)) == 0) {
            // Left (Back)
            MENU_updatestate(MENU_INPUT_BACK);
        } else if ((PINB & _BV(PB2)) == 0) {
            // Down
            MENU_updatestate(MENU_INPUT_DOWN);
        } else if ((PINB & _BV(PB1)) == 0) {
            // Up
            MENU_updatestate(MENU_INPUT_UP);
        } else if ((PINB & _BV(PB3)) == 0) {
            MENU_updatestate(MENU_INPUT_OK);
        }
        ui_check = 0;
    }
}


void loop()
{
    for(;;) {
        // sprintf(buffer, "out=%d ", output_dac);
        // LCD_printAt(LCD_INSTR_nextLine, buffer);
    }
}

void SignalInput_init()
{
     // Initialize ADC
    ADMUX = 0
        // Internal 2.56V Voltage Reference with external capacitor at AREF pin
        | _BV(REFS1)
        | _BV(REFS0)

        // Measure ADC1 - ADC0 with 10x gain
        | 0b01001;

    ADCSRA = 0
        // Activate ADC interrupt
        | _BV(ADIE)

        // Activate ADC Auto Triggering
        | _BV(ADATE)

        // TODO: try more prescaler settings
        // Set ADC Prescaler to 64 => f(ADC) = 16MHz / 64 = 250 kHz
        | 0b110;

    // ADC Free Running Mode - Auto Trigger
    ADCSRB = 0;

    // Disable digital input buffer for ADC pins (ADC0, ADC1)
    DIDR0 = 0b00000011;

    // Enable ADC
    ADCSRA |= _BV(ADEN);
    
    // First ADC conversion
    ADCSRA |= _BV(ADSC);
}

void SignalOutput_init()
{
    // initialzie external DAC
    TWI_init();

    // initialize timer1 to be used with the tremolo effect
    TCCR1A = 0;
    TCCR1B = 0
        | _BV(WGM12)
        | _BV(CS12)
        | _BV(CS10);
    
    // TODO settings
    OCR1A = TIMER_1HZ / 2;

    //initialize timer0 to be used with the flanger effect
    // TODO settings
    timer0_top = TIMER_1HZ;
    TCCR0A = 0
        | _BV(WGM01);
    TCCR0B = 0
        | _BV(CS02)
        | _BV(CS00);
    TIMSK0 = 0
        | _BV(OCIE0A);
    OCR0A = 0xFF;
}

void UI_init()
{
     // Activate pull-up resistors for control buttons
    PORTB |= _BV(PB0) | _BV(PB1) | _BV(PB2) | _BV(PB3);

    // Activate interrupts for PB0..3
    PCICR  = (1 << PCIE1);
    PCMSK1 = (1 << PCINT8) | (1 << PCINT9)
        | (1 << PCINT10) | (1 << PCINT11);
}

int main()
{
    // Activate global interrupts
    sei();

    /* Initialize */
    MENU_init();
    UI_init();
    SignalOutput_init();
    SignalInput_init();

    /* Main Loop */
    loop();

    return 0;
}
